# ========================================================================
# TestOneHotEncoding04.py: ...
# 
# Note: Still need to check that the onehot encoding for the catergory 
# is working correctly ... 
# 
# Author: pavankalyan
# Modified by: Mark Austin                                   December 2021 
# ========================================================================

import numpy as np
import matplotlib.pyplot as plt
import pandas as pd

from sklearn.metrics import confusion_matrix
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import LabelEncoder, OneHotEncoder
from sklearn.preprocessing import StandardScaler
from sklearn.compose import ColumnTransformer

# ========================================================================
# Useful activation functions and their derivatives ...
# ========================================================================

def sigmoid_(Z):
    return 1/(1+np.exp(-Z))

def relu_(Z):
    return Z*(Z>0)

def drelu_(Z):
    return 1. *(Z>0)

def dsigmoid_(Z):
    return sigmoid_(Z)*(1-sigmoid_(Z))

def sigmoid(Z):
    return sigmoid_(Z),Z

def relu(Z):
    return relu_(Z),Z

def linear_activation_forward(A_prev,W,b,activation):
    if activation == "sigmoid":
        Z, linear_cache = linear_forward(A_prev,W,b)
        A, activation_cache = sigmoid(Z)

    elif activation == "relu":
        Z, linear_cache = linear_forward(A_prev,W,b)
        A, activation_cache = relu(Z)

    assert (A.shape == (W.shape[0], A_prev.shape[1]))
    cache = (linear_cache, activation_cache)

    return A, cache

# ========================================================================
# initialize parameters ...
# ========================================================================

def initialize_parameters(n_x, n_h, n_y):
    np.random.seed(1)

    W1 = np.random.randn(n_h,n_x)*0.01
    b1 = np.zeros((n_h,1))
    W2 = np.random.randn(n_y,n_h)*0.01
    b2 = np.zeros((n_y,1))

    # assert(W1.shape == (n_h, n_x))
    # assert(b1.shape == (n_h, 1))
    # assert(W2.shape == (n_y, n_h))
    # assert(b2.shape == (n_y, 1))

    parameters = {"W1": W1,
                  "b1": b1,
                  "W2": W2,
                  "b2": b2}

    return parameters

# ========================================================================
# initialize parameters (deep layer) ...
# ========================================================================

def initialize_parameters_deep(layer_dims):
    np.random.seed(3)
    parameters = {}
    L = len(layer_dims)
    for l in range(1, L):
        parameters['W' + str(l)] = np.random.randn(layer_dims[l],layer_dims[l-1])*0.01
        parameters['b' + str(l)] = np.zeros((layer_dims[l],1))

        assert(parameters['W' + str(l)].shape == (layer_dims[l], layer_dims[l-1]))
        assert(parameters['b' + str(l)].shape == (layer_dims[l], 1))
    return parameters

def linear_forward(A, W, b):
    Z = np.dot(W,A)+b
    assert(Z.shape == (W.shape[0], A.shape[1]))
    cache = (A, W, b)

    return Z, cache

# ========================================================================
# forward propogation for L-layer neural network ...
# ========================================================================

def L_model_forward(X, parameters):
    caches = []
    A = X
    L = len(parameters) // 2
    for l in range(1, L):
        A_prev = A
        A, cache = linear_activation_forward(A_prev,parameters['W'+str(l)],parameters['b'+str(l)],"relu")
        caches.append(cache)
    AL, cache = linear_activation_forward(A,parameters['W'+str(L)],parameters['b'+str(L)],"sigmoid")
    caches.append(cache)
    assert(AL.shape == (1,X.shape[1]))
    return AL, caches

# ========================================================================
# cost function ...
# ========================================================================

def compute_cost(AL, Y):
    m=Y.shape[1]
    cost = -(1/m)*np.sum((Y*np.log(AL)+(1-Y)*np.log(1-AL)))
    cost=np.squeeze(cost)
    assert(cost.shape == ())
    return cost

# ========================================================================
# back propogation for one layer neural network ...
# ========================================================================

def linear_backward(dZ, cache):
    A_prev, W, b = cache
    m = A_prev.shape[1]
    dW = (1/m)*np.dot(dZ,A_prev.T)
    db = (1/m)*np.sum(dZ,axis=1,keepdims=True)
    dA_prev = np.dot(W.T,dZ)

    assert (dA_prev.shape == A_prev.shape)
    assert (dW.shape == W.shape)
    assert (db.shape == b.shape)

    return dA_prev, dW, db

def relu_backward(dA,activation_cache):
    return dA* drelu_(activation_cache)

def sigmoid_backward(dA,activation_cache):
    return dA* dsigmoid_(activation_cache)

def linear_activation_backward(dA, cache, activation):
    linear_cache, activation_cache = cache
    if activation == "relu":
        dZ = relu_backward(dA,activation_cache)
        dA_prev, dW, db = linear_backward(dZ,linear_cache)

    elif activation == "sigmoid":
        dZ = sigmoid_backward(dA,activation_cache)
        dA_prev, dW, db = linear_backward(dZ,linear_cache)
    return dA_prev,dW,db

# ========================================================================
# Back propogation for L layers ...
# ========================================================================

def L_model_backward(AL, Y, caches):
    grads = {}
    L = len(caches)
    m = AL.shape[1]
    Y = Y.reshape(AL.shape)

    dAL = - (np.divide(Y, AL) - np.divide(1 - Y, 1 - AL))

    current_cache = caches[L-1]
    grads["dA" + str(L-1)], grads["dW" + str(L)], grads["db" + str(L)] = linear_activation_backward(dAL,current_cache,"sigmoid")

    for l in reversed(range(L-1)):
        current_cache = caches[l]
        dA_prev_temp, dW_temp, db_temp = linear_activation_backward(grads["dA"+str(l+1)],current_cache,"relu")
        grads["dA" + str(l)] = dA_prev_temp
        grads["dW" + str(l + 1)] = dW_temp
        grads["db" + str(l + 1)] = db_temp
    return grads

# ========================================================================
# Update parameters ...
# ========================================================================

def update_parameters(parameters, grads, learning_rate):
    L = len(parameters) // 2
    for l in range(L):
        parameters["W" + str(l+1)] = parameters["W" + str(l+1)]-(learning_rate)*grads["dW"+str(l+1)]
        parameters["b" + str(l+1)] = parameters["b" + str(l+1)]-(learning_rate)*grads["db"+str(l+1)]
    return parameters


# ========================================================================
# L-layer model ...
# ========================================================================

def L_layer_model(X, Y, layers_dims, learning_rate = 0.0075, num_iterations = 3000, print_cost=False):
    np.random.seed(1)
    costs = []

    parameters = initialize_parameters_deep(layers_dims)

    for i in range(0, num_iterations):
        AL, caches = L_model_forward(X, parameters)
        cost = compute_cost(AL, Y)
        grads = L_model_backward(AL, Y, caches)
        parameters = update_parameters(parameters, grads, learning_rate)
        if print_cost and i % 100 == 0:
            print ("Cost after iteration %i: %f" %(i, cost))
        if print_cost and i % 100 == 0:
            costs.append(cost)
    # plot the cost
    plt.plot(np.squeeze(costs))
    plt.ylabel('cost')
    plt.xlabel('iterations (per tens)')
    plt.title("Learning rate =" + str(learning_rate))
    plt.show()

    return parameters

# ========================================================================
# Two layer model ...
# ========================================================================

def two_layer_model(X, Y, layers_dims, learning_rate = 0.0075, num_iterations = 3000, print_cost=False):
    np.random.seed(1)
    grads = {}
    costs = []
    m = X.shape[1]
    (n_x, n_h, n_y) = layers_dims

    parameters = initialize_parameters(n_x,n_h,n_y)

    W1 = parameters["W1"]
    b1 = parameters["b1"]
    W2 = parameters["W2"]
    b2 = parameters["b2"]

    for i in range(0, num_iterations):
        A1, cache1 = linear_activation_forward(X,W1,b1,"relu")
        A2, cache2 = linear_activation_forward(A1,W2,b2,"sigmoid")

        cost = compute_cost(A2, Y)

        dA2 = - (np.divide(Y, A2) - np.divide(1 - Y, 1 - A2))

        dA1, dW2, db2 = linear_activation_backward(dA2, cache2, "sigmoid")
        dA2 = - (np.divide(Y, A2) - np.divide(1 - Y, 1 - A2))

        dA1, dW2, db2 = linear_activation_backward(dA2, cache2, "sigmoid")
        dA0, dW1, db1 = linear_activation_backward(dA1, cache1, "relu")

        grads['dW1'] = dW1
        grads['db1'] = db1
        grads['dW2'] = dW2
        grads['db2'] = db2

        parameters = update_parameters(parameters, grads, learning_rate)

        W1 = parameters["W1"]
        b1 = parameters["b1"]
        W2 = parameters["W2"]
        b2 = parameters["b2"]

        if print_cost and i % 100 == 0:
            print("Cost after iteration {}: {}".format(i, np.squeeze(cost)))
        if print_cost and i % 100 == 0:
            costs.append(cost)
    plt.plot(np.squeeze(costs))
    plt.ylabel('cost')
    plt.xlabel('iterations (per tens)')
    plt.title("Learning rate =" + str(learning_rate))
    plt.show()

    return parameters

# ========================================================================
# predictions ...
# ========================================================================

def predict(X,Y,parameters):
    W1=parameters["W1"]
    b1=parameters["b1"]
    W2=parameters["W2"]
    b2=parameters["b2"]
    A1, cache1 = linear_activation_forward(X,W1,b1,"relu")
    A2, cache2 = linear_activation_forward(A1,W2,b2,"sigmoid")
    return (A2>=0.5)*1

def predict_L_layer(X,Y,parameters):
    AL,caches=L_model_forward(X,parameters)
    return (AL>=0.5)*1

# ========================================================================
# main method ...
# ========================================================================

def main():
    print("--- Enter TestOneHotEncoding04.main()        ... ");
    print("--- ======================================== ... ");

    # Part 1: Import dataset ...

    print("--- import dataset ... ");

    dataset = pd.read_csv('../data/Churn_Modelling.csv')
    X = dataset.iloc[:, 3:13].values
    y = dataset.iloc[:, 13].values

    print( dataset.head() )

    # Part 2: encode categorical data ...

    print("--- encode categorical data ... ");

    labelencoder_X_1 = LabelEncoder()
    X[:, 1] = labelencoder_X_1.fit_transform(X[:, 1])
    labelencoder_X_2 = LabelEncoder()
    X[:, 2] = labelencoder_X_2.fit_transform(X[:, 2])

    # Following method has been deprecated -- need to fix ...

    # onehotencoder = OneHotEncoder(categorical_features = [1] )
    # onehotencoder = OneHotEncoder(categorical_features = 'all')

    onehotencoder = ColumnTransformer([("Surname", OneHotEncoder(), [0])],
                                      remainder = "passthrough")

    X = onehotencoder.fit_transform(X).toarray()
    print (X)

    # To avoid dummy variable trap

    X = X[:, 1:]

    # Part 3: Split data into train and test sets ...

    print("--- split data into train and test sets ... ");

    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size = 0.2, random_state = 0)

    # Part 4: Feature scaling ...

    print("--- feature scaling ... ");

    sc = StandardScaler()
    X_train = sc.fit_transform(X_train)
    X_test  = sc.transform(X_test)

    X_train = X_train.T
    X_test  = X_test.T
    y_train = y_train.reshape(y_train.shape[0],1)
    y_test  = y_test.reshape(y_test.shape[0],1)
    y_train = y_train.T
    y_test  = y_test.T

    # Part 5: Initialize parameters (W and b) ...

    print("--- Initialize parameters (W and b) ... ");

    n_x=X_train.shape[0]
    n_h=10
    n_y=y_train.shape[0]

    parameters_=initialize_parameters(n_x,n_h,n_y)

    print("W1 = " + str(parameters_["W1"]))
    print("b1 = " + str(parameters_["b1"]))
    print("W2 = " + str(parameters_["W2"]))
    print("b2 = " + str(parameters_["b2"]))
    print(n_x)

    print(parameters_["W1"].shape)
    print(parameters_["b1"].shape)
    print(parameters_["W2"].shape)
    print(parameters_["b2"].shape)

    # Part 6: Initialize parameters for deep neural network ...

    print("--- Initialize parameters for deep neural network ...");

    parameters=initialize_parameters_deep([n_x,10,10,n_y])
    W1=parameters_["W1"]
    b1=parameters_["b1"]

    A=X_train
    Z, linear_cache = linear_forward(A, W1, b1)

    A, linear_activation_cache = linear_activation_forward(X_train, W1, b1, activation = "sigmoid")
    print("--- With Sigmoid A:"+ str(A))

    A, linear_activation_cache = linear_activation_forward(X_train, W1, b1, activation = "relu")
    print("--- With RELU A:"+ str(A))

    AL,caches=L_model_forward(X_train,parameters)
    print("--- AL = " + str(AL))
    print("--- Length of caches list = " + str(len(caches)))

    print("--- Cost : "+str(compute_cost(AL,y_train)))

    dZ=sigmoid_(Z)
    dA_prev,dW,db=linear_backward(dZ,linear_cache)
    print ("dA_prev = "+ str(dA_prev.shape))
    print ("dW = " + str(dW.shape))
    print ("db = " + str(db.shape))

    dAL=AL-y_train

    dA_prev, dW, db = linear_activation_backward(dAL, linear_activation_cache, activation = "sigmoid")
    print ("sigmoid:")
    print ("dA_prev = "+ str(dA_prev.shape))
    print ("dW = " + str(dW.shape))
    print ("db = " + str(db.shape) + "\n")

    dA_prev, dW, db = linear_activation_backward(dAL, linear_activation_cache, activation = "relu")
    print ("relu:")
    print ("dA_prev = "+ str(dA_prev.shape))
    print ("dW = " + str(dW.shape))
    print ("db = " + str(db.shape) + "\n")

    # Part 7: Back propogation for L layers ...

    print("--- Back propogation for L layers ... ");

    grads=L_model_backward(AL,y_train,caches)

    # Part 8: Update parameters ...

    print("--- Update parameters ... ");

    parameters = update_parameters(parameters, grads, 0.1)

    print ("W1 = "+ str(parameters["W1"]))
    print ("b1 = "+ str(parameters["b1"]))
    print ("W2 = "+ str(parameters["W2"]))
    print ("b2 = "+ str(parameters["b2"]))

    # Part 9: Run two-layer model ...

    print("--- Run two-layer model ... ");

    parameters = two_layer_model(X_train, y_train, layers_dims = (n_x, n_h, n_y), num_iterations = 10000, print_cost=True)

    # Part 10: Compute predicitions ...

    print("--- Compute predictions ... ");

    predictions_train = predict(X_train,y_train, parameters)

    from sklearn.metrics import confusion_matrix
    cm_train=confusion_matrix(np.squeeze(y_train.T),np.squeeze(predictions_train.T))

    predictions_test=predict(X_test,y_test,parameters)
    cm_test=confusion_matrix(np.squeeze(y_test.T),np.squeeze(predictions_test.T))

    # Part 11: N layer neural network

    print("--- Run N layer neural network Compute predictions ... ");

    layers_dims=[n_x,1,1,1]
    parameters_L = L_layer_model(X_train, y_train, layers_dims, num_iterations = 10000, print_cost = True)

    predictions_train_L = predict_L_layer(X_train,y_train, parameters_L)
    cm_train_L=confusion_matrix(np.squeeze(y_train.T),np.squeeze(predictions_train_L.T))

    predictions_test_L=predict_L_layer(X_test,y_test,parameters_L)
    cm_test_L=confusion_matrix(np.squeeze(y_test.T),np.squeeze(predictions_test_L.T))

    print((cm_train_L[0][0]+cm_train_L[1][1])/8000)
    print((cm_test_L[0][0]+cm_test_L[1][1])/2000)

    print("--- ======================================== ... ");
    print("--- Leave TestOneHotEncoding04.main()        ... ");

# call the main method ...

main()
