# ========================================================================
# TestRecurrentNeuralNetwork01.py: Simplified LSTM model for prediction of
#                                  shampoo sales.
# 
# Dataset: 3 years of sales for shampoo. The first two years are allocated
# to training. Year 3 is used for testing of actual sales vs. predicted
# sales.
# 
# Developed May 2022. 
# ========================================================================

from pandas import DataFrame
from pandas import Series
from pandas import concat
from pandas import read_csv
from pandas import datetime
from sklearn.metrics import mean_squared_error
from sklearn.preprocessing import MinMaxScaler
from keras.models import Sequential
from keras.layers import Dense
from keras.layers import LSTM
from math import sqrt
from matplotlib import pyplot
import numpy

# ==================================================
# date-time parsing function for loading the dataset
# ==================================================

def parser(x):
    return datetime.strptime('190'+x, '%Y-%m')

# ==================================================
# frame a sequence as a supervised learning problem
# ==================================================

def timeseries_to_supervised(data, lag=1):
    df = DataFrame(data)
    columns = [df.shift(i) for i in range(1, lag+1)]
    columns.append(df)
    df = concat(columns, axis=1)
    df.fillna(0, inplace=True)
    return df
 
# ==================================================
# create a differenced series
# ==================================================

def difference(dataset, interval=1):
    diff = list()
    for i in range(interval, len(dataset)):
       value = dataset[i] - dataset[i - interval]
       diff.append(value)
    return Series(diff)
 
# ==================================================
# invert differenced value
# ==================================================

def inverse_difference(history, yhat, interval=1):
    return yhat + history[-interval]
 
# ==================================================
# scale train and test data to [-1, 1]
# ==================================================

def scale(train, test):

    # fit scaler

    scaler = MinMaxScaler(feature_range=(-1, 1))
    scaler = scaler.fit(train)

    # transform train

    train = train.reshape(train.shape[0], train.shape[1])
    train_scaled = scaler.transform(train)

    # transform test

    test = test.reshape(test.shape[0], test.shape[1])
    test_scaled = scaler.transform(test)
    return scaler, train_scaled, test_scaled
 
# ==================================================
# inverse scaling for a forecasted value
# ==================================================

def invert_scale(scaler, X, value):
    new_row = [x for x in X] + [value]
    array = numpy.array(new_row)
    array = array.reshape(1, len(array))
    inverted = scaler.inverse_transform(array)
    return inverted[0, -1]
 
# ==================================================
# fit an LSTM network to training data
# ==================================================

def fit_lstm(train, batch_size, nb_epoch, neurons):

    print("--- Enter fit_lstm()                     ... ");
    print("--- ==================================== ... ");

    X, y = train[:, 0:-1], train[:, -1]
    X = X.reshape(X.shape[0], 1, X.shape[1])

    print("--- data array train ...");
    print("--- train.shape:  ", train.shape );

    print(train)

    print("--- data array X ...");
    print("--- X.shape:  ", X.shape );

    print(X)

    print("--- data array y ...");
    print("--- y.shape:  ", y.shape );

    print(y)

    print("--- Use sequential Keras API to create LSTM network ...");

    model = Sequential()
    model.add(LSTM(neurons, batch_input_shape=(batch_size, X.shape[1], X.shape[2]), stateful=True))
    model.add(Dense(1))
    model.compile(loss='mean_squared_error', optimizer='adam')

    print("--- Fit model to data ... ");

    for i in range(nb_epoch):
      model.fit(X, y, epochs=1, batch_size=batch_size, verbose=0, shuffle=False)
      model.reset_states()

    print("--- ==================================== ... ");
    print("--- Leave fit_lstm()                     ... ");

    return model
 
# ==================================================
# make a one-step forecast
# ==================================================

def forecast_lstm(model, batch_size, X):
    X = X.reshape(1, 1, len(X))
    yhat = model.predict(X, batch_size=batch_size)
    return yhat[0,0]

# ==============================
# main method ...
# ==============================

def main():
    print("--- Enter TestRecurrentNeuralNetwork01.main() ... ");
    print("--- ========================================= ... ");

    # load dataset

    print("--- Part 01: Load and print shampoo sales dataset ... ");

    series = read_csv('../data/shampoo-sales.csv', header=0,
                      parse_dates=[0], index_col=0, squeeze=True, date_parser=parser)

    # summarize first few rows

    print("--- Series header ... ");

    print(series.head())

    # Iterate over series rows ...

    print("--- Iterate over series rows ... ");

    for ind, val in series.iteritems():
       print( ind, val )

    # line plot

    print("--- Part 02: Create series plot ... ");

    series.plot()
    pyplot.show()
 
    # transform data to be stationary

    print("--- Part 03: Transform data to be stationary ... ");

    raw_values  = series.values
    diff_values = difference(raw_values, 1)
 
    # transform data to be supervised learning

    print("--- Part 04: Transform data to be supervised learning ... ");

    supervised = timeseries_to_supervised(diff_values, 1)
    supervised_values = supervised.values
 
    # split data into train and test-sets

    print("--- Part 05: split data into train and testing datasets ... ");

    train, test = supervised_values[0:-12], supervised_values[-12:]

    print("--- train.shape: ", train.shape );
    print("--- test.shape:  ", test.shape );
 
    # transform the scale of the data

    print("--- Part 06: transform scale of the data ... ");

    scaler, train_scaled, test_scaled = scale(train, test)

    # fit the model

    print("--- Part 07: Fit the model (nepoch = 2000; network size = 6) ... ");

    lstm_model = fit_lstm(train_scaled, 1, 2000, 6)

    # forecast the entire training dataset to build up state for forecasting

    print("--- Part 08: Forecast the entire training dataset ... ");

    train_reshaped = train_scaled[:, 0].reshape(len(train_scaled), 1, 1)
    lstm_model.predict(train_reshaped, batch_size=1)
 
    # walk-forward validation on the test data

    print("--- Part 09: Validate model on test data ... ");
    print("--- ");

    predictions = list()
    for i in range(len(test_scaled)):
       # make one-step forecast
       X, y = test_scaled[i, 0:-1], test_scaled[i, -1]
       yhat = forecast_lstm(lstm_model, 1, X)
       # invert scaling
       yhat = invert_scale(scaler, X, yhat)
       # invert differencing
       yhat = inverse_difference(raw_values, yhat, len(test_scaled)+1-i)
       # store forecast
       predictions.append(yhat)
       expected = raw_values[len(train) + i + 1]
       print('--- Month=%d, Predicted=%f, Expected=%f' % (i+1, yhat, expected))
 
    # report performance

    print("--- Part 10: Report performance ... ");

    rmse = sqrt(mean_squared_error(raw_values[-12:], predictions))
    print('Test RMSE: %.3f' % rmse)

    # line plot of observed vs predicted

    print("--- Part 11: Line plot of observed vs predicted ... ");

    pyplot.plot(raw_values[-12:])
    pyplot.plot(predictions)
    pyplot.legend(fontsize=12)
    pyplot.xlabel("time")
    pyplot.ylabel("shampoo price")
    pyplot.grid(True)

    pyplot.show()
 
    print("--- ========================================= ... ");
    print("--- Leave TestRecurrentNeuralNetwork01.main() ... ");

# ==================================================
# call the main method ...
# ==================================================

main()
