# ========================================================================
# TestRecurrentNeuralNetworkStocks.py.
# 
# Written by: Mark Austin
# ========================================================================

from pandas import DataFrame
from pandas import Series
from pandas import concat
from pandas import read_csv
from pandas import datetime
from sklearn.metrics import mean_squared_error
from sklearn.preprocessing import MinMaxScaler
from keras.models import Sequential
from keras.layers import Dense
from keras.layers import LSTM
from math import sqrt
from matplotlib import pyplot

import numpy as np 
import pandas as pd

import matplotlib.pyplot as plt

# ==================================================
# date-time parsing function for loading the dataset
# ==================================================

def parser(x):
    return datetime.strptime('190'+x, '%Y-%m')

# ==================================================
# frame a sequence as a supervised learning problem
# ==================================================

def timeseries_to_supervised(data, lag=1):
    df = DataFrame(data)
    columns = [df.shift(i) for i in range(1, lag+1)]
    columns.append(df)
    df = concat(columns, axis=1)
    df.fillna(0, inplace=True)
    return df
 
# ==================================================
# create a differenced series
# ==================================================

def difference(dataset, interval=1):

    print("--- Enter difference() ... ");
    print("--- interval = %d ...\n" % (interval) );

    print(dataset);

    diff = list()
    for i in range(interval, len(dataset)):
       value = dataset[i] - dataset[i - interval]
       diff.append(value)

    print("--- List of differences ... ");
    print(diff);
    print("--- Leaving difference() ... ");

    return Series(diff)
 
# ==================================================
# invert differenced value
# ==================================================

def inverse_difference(history, yhat, interval=1):
    return yhat + history[-interval]
 
# ==================================================
# scale train and test data to [-1, 1]
# ==================================================

def scale(train, test):

    # fit scaler

    scaler = MinMaxScaler(feature_range=(-1, 1))
    scaler = scaler.fit(train)

    # transform train

    train = train.reshape(train.shape[0], train.shape[1])
    train_scaled = scaler.transform(train)

    # transform test

    test = test.reshape(test.shape[0], test.shape[1])
    test_scaled = scaler.transform(test)
    return scaler, train_scaled, test_scaled
 
# ==================================================
# inverse scaling for a forecasted value
# ==================================================

def invert_scale(scaler, X, value):
    new_row = [x for x in X] + [value]
    array = numpy.array(new_row)
    array = array.reshape(1, len(array))
    inverted = scaler.inverse_transform(array)
    return inverted[0, -1]
 
# ==================================================
# fit an LSTM network to training data
# ==================================================

def fit_lstm(train, batch_size, nb_epoch, neurons):

    print("--- Enter fit_lstm()                     ... ");
    print("--- ==================================== ... ");

    X, y = train[:, 0:-1], train[:, -1]
    X = X.reshape(X.shape[0], 1, X.shape[1])

    print("--- data array train ...");
    print("--- train.shape:  ", train.shape );

    print(train)

    print("--- data array X ...");
    print("--- X.shape:  ", X.shape );

    print(X)

    print("--- data array y ...");
    print("--- y.shape:  ", y.shape );

    print(y)

    print("--- Use sequential Keras API to create LSTM network ...");

    model = Sequential()
    model.add(LSTM(neurons, batch_input_shape=(batch_size, X.shape[1], X.shape[2]), stateful=True))
    model.add(Dense(1))
    model.compile(loss='mean_squared_error', optimizer='adam')

    print("--- Fit model to data ... ");

    for i in range(nb_epoch):
      model.fit(X, y, epochs=1, batch_size=batch_size, verbose=0, shuffle=False)
      model.reset_states()

    print("--- ==================================== ... ");
    print("--- Leave fit_lstm()                     ... ");

    return model
 
# ==================================================
# make a one-step forecast
# ==================================================

def forecast_lstm(model, batch_size, X):
    X = X.reshape(1, 1, len(X))
    yhat = model.predict(X, batch_size=batch_size)
    return yhat[0,0]

# ==============================
# main method ...
# ==============================

def main():
    print("--- Enter TestStocks.main()                  ... ");
    print("--- ======================================== ... ");

    # load dataset

    print("--- Load AAPL, MSFT, GOOG stock price dataset ... ");

    df = pd.read_csv("../data/stocks.csv")

    # print dataframe contents ...

    print("--- Dataframe contents ... ");

    print(df)

    # Filter dataframe to keep only the AAPL entries ...
    # selecting rows based on condition 
    # Assign index numbers to 0, 1, ... 122.

    print("--- Filter dataframe to keep only AAPL entries ... \n")

    options = ['AAPL'] 
    df_aapl = df [ df['symbol'].isin(options) ].copy()  
    df_aapl.index = list( range(0, 123, 1 ) )
    
    print('--- AAPL dataframe :\n', df_aapl)

    # =============================================================
    # Create line plots: MSFT, GOOG and AAPL price vs date.
    # =============================================================

    ax = plt.gca()
    df_aapl.plot(kind='line',x='date',y='price',color='blue',ax=ax)
    plt.show()
 
    # Transform data to be stationary

    print("--- Extract list of raw values from AAPL data ... ");

    raw_values  = pd.Series( data=df_aapl.price )
    print( raw_values )

    print("--- Transform AAPL data to be stationary ... ");

    diff_values = difference(raw_values, 1)
 
    # transform data to be supervised learning

    print("--- Transform data to be supervised learning ... ");

    supervised = timeseries_to_supervised(diff_values, 1)
    supervised_values = supervised.values
 
    # split data into train and test-sets

    print("--- split data into train and testing datasets ... ");

    train, test = supervised_values[0:-12], supervised_values[-12:]

    print("--- train.shape: ", train.shape );
    print("--- test.shape:  ", test.shape );
 
    # transform the scale of the data

    print("--- transform scale of the data ... ");

    scaler, train_scaled, test_scaled = scale(train, test)

    # fit the model

    print("--- fit the model (nepoch = 2000; network size = 6) ... ");

    lstm_model = fit_lstm(train_scaled, 1, 2000, 6)

    # forecast the entire training dataset to build up state for forecasting

    print("--- forecast the entire training dataset ... ");

    train_reshaped = train_scaled[:, 0].reshape(len(train_scaled), 1, 1)
    lstm_model.predict(train_reshaped, batch_size=1)
 
    # walk-forward validation on the test data

    print("--- validate model on test data ... ");
    print("--- ");

    predictions = list()
    for i in range(len(test_scaled)):
       # make one-step forecast
       X, y = test_scaled[i, 0:-1], test_scaled[i, -1]
       yhat = forecast_lstm(lstm_model, 1, X)
       # invert scaling
       yhat = invert_scale(scaler, X, yhat)
       # invert differencing
       yhat = inverse_difference(raw_values, yhat, len(test_scaled)+1-i)
       # store forecast
       predictions.append(yhat)
       expected = raw_values[len(train) + i + 1]
       print('--- Month=%d, Predicted=%f, Expected=%f' % (i+1, yhat, expected))
 
    # report performance

    print("--- report performance ... ");

    rmse = sqrt(mean_squared_error(raw_values[-12:], predictions))
    print('Test RMSE: %.3f' % rmse)

    # line plot of observed vs predicted

    print("--- line plot of observed vs predicted ... ");

    pyplot.plot(raw_values[-12:])
    pyplot.plot(predictions)
    pyplot.legend(fontsize=12)
    pyplot.xlabel("time")
    pyplot.ylabel("shampoo price")
    pyplot.grid(True)

    pyplot.show()
 
    print("--- ======================================== ... ");
    print("--- Leave TestRNNShampooSales.main()         ... ");

# ==================================================
# call the main method ...
# ==================================================

main()
