# ===============================================================
# TestMultiLayer-Virus-Contraction.py: Predict virus contraction.
# 
# Medical data:
# 
# -------------------------------------------------------
#            Loss of   Weight    Runny    Body
#    Person    Smell     Loss     Nose    Pain  Positive?
# -------------------------------------------------------
#         1        1        0        0       1          1
#         2        1        0        0       0          1
#         3        0        0        1       1          0
#         4        0        1        0       0          0
#         5        1        1        0       0          1
#         6        0        0        1       1          1
#         7        0        0        0       1          0
#         8        0        0        1       0          0
# -------------------------------------------------------
# 
# Note: There are two instances of [ 0, 0, 1, 1 ] input.
# One led to a positive result, one to a negative result. 
# 
# Reference: Shukla, et al., Neural Networks from Scratch with
# Python Code and Math in Detail, Towards AI, 2020  
# 
# Network Architecture:
# --------------------------------------------------------------- 
# 
# 2 input nodes, 4 nodes on hidden layer, 1 node on output. 
# 
# Weighted edges: 
# 
# input  layer --> hidden layer: connected by 4x4 = 16 weights. 
# hidden layer --> output layer: connected by 4 weights. 
# 
# Reference: Shukla, et al., Neural Networks from Scratch with
# Python Code and Math in Detail, Towards AI, 2020  
# 
# Modified by: Mark Austin                          October, 2020 
# ===============================================================

import math
import matplotlib
import matplotlib.pyplot as plt
import numpy as np

# Define Sigmoid function:

def sigmoid(x):
   return 1/(1+np.exp(-x))

# Define derivative of Sigmoid function:

def sigmoid_der(x):
   return sigmoid(x)*(1-sigmoid(x))

# main method ...

def main():
    print("--- Enter TestMultiLayer-Virus-Contraction.main() ... ");
    print("--- ============================================= ... ");

    print("--- ");
    print("--- Part 1: Define input features ... ");

    input_features = np.array( [ [1,0,0,1],
                                 [1,0,0,0],
                                 [0,0,1,1],
                                 [0,1,0,0],
                                 [1,1,0,0],
                                 [0,0,1,1],
                                 [0,0,0,1],
                                 [0,0,1,0] ] )

    print (input_features.shape)
    print (input_features)

    # Define target output:

    print("--- ");
    print("--- Part 2: Define target output ... ");

    target_output = np.array([[1,1,0,0,1,1,0,0]])

    # Reshaping target output into vector:

    target_output = target_output.reshape(8,1)
    print (target_output)

    print("--- ");
    print("--- Part 3: Initial weights and learning parameters ... ");
    print("--- ");
    
    weight_hidden = np.random.rand(4,4)
    weight_output = np.random.rand(4,1)

    print(weight_hidden)
    print(weight_output)

    bias = 0.3       # Bias weight: 
    lr   = 0.10      # Learning Rate:

    # ========================================================
    # Main logic for neural network:
    # ========================================================

    print("--- ");
    print("--- Part 4: Train Neural Network ... ");
    print("--- ---------------------------- ... ");
    print("--- ");

    # Running our code 20000 times:

    for epoch in range(20000):

       print("--- Start epoch: %f ..." %(epoch))

       # Input for hidden layer :

       print("---   Step 1: Compute input for hidden layer ...")

       input_hidden = np.dot(input_features, weight_hidden)
 
       # Output from hidden layer :

       print("---   Step 2: Compute output from hidden layer ...")

       output_hidden = sigmoid(input_hidden)
 
       # Input for output layer :

       print("---   Step 3: Compute input for output layer ...")

       input_op = np.dot(output_hidden, weight_output)
 
       # Output from output layer :

       print("---   Step 4: Compute output for output layer ...")

       output_op = sigmoid(input_op)

       # ======================================================
       # Phase 1: Compute derivatives for phase 1 ...
       # ======================================================
 
       # Calculating Mean Squared Error :

       print("---   Step 5: Compute mean square error ...")

       error_out = ((1 / 2) * (np.power((output_op - target_output), 2)))

       print("---   Mean square error: %f ... \n" %( error_out.sum() ) )
 
       # Derivatives for phase 1 :

       print("---   Step 6: Compute derivatives for Phase 1 ...")

       derror_douto = output_op - target_output
       douto_dino   = sigmoid_der(input_op) 
       dino_dwo     = output_hidden

       derror_dwo = np.dot(dino_dwo.T, derror_douto * douto_dino)

       # ======================================================
       # Phase 2:
       # 
       # derror_w1    = derror_douth * douth_dinh * dinh_dw1
       # derror_douth = derror_dino * dino_outh
       # ======================================================
 
       # Derivatives for phase 2 :

       print("---   Step 7: Compute derivatives for Phase 2 ...")

       derror_dino  = derror_douto * douto_dino
       dino_douth   = weight_output
       derror_douth = np.dot(derror_dino , dino_douth.T)

       douth_dinh = sigmoid_der(input_hidden) 
       dinh_dwh   = input_features
       derror_dwh = np.dot(dinh_dwh.T, douth_dinh * derror_douth)

       print("---   Step 8: Update weights in hidden and output layers ...")

       weight_hidden -= lr * derror_dwh
       weight_output -= lr * derror_dwo

    # ============================================
    # Print Results ...
    # ============================================

    print("--- ");
    print("--- ------------------------------------- ... ");
    print("--- Part 5: Summary of weights            ... ");
    print("--- ------------------------------------- ... ");
    print("--- Final weights in hidden layer:");
    print("--- ");

    print (weight_hidden) 

    print("--- ");
    print("--- Final weights in output layer:");
    print("--- ");

    print (weight_output)

    print("--- ");
    print("--- --------------------------------------------- ... ");
    print("--- Part 6: Use trained network to predict values ... ");
    print("--- --------------------------------------------- ... ");

    print("--- ");
    print("--- Part 6.1: Verify input [1,0,0,1] --> 1 ... ");
    print("--- ");

    single_point = np.array([1,0,0,1]) #1st step:

    result1 = np.dot(single_point, weight_hidden) 
    result2 = sigmoid(result1)
    result3 = np.dot( result2, weight_output)
    result4 = sigmoid(result3)

    print ("--- Result: %f ..." %(result4))

    print("--- ");
    print("--- Part 6.2: Verify input [1,0,0,0] --> 1 ... ");
    print("--- ");

    single_point = np.array([1,0,0,0]) #1st step:

    result1 = np.dot(single_point, weight_hidden) 
    result2 = sigmoid(result1)
    result3 = np.dot(result2, weight_output)
    result4 = sigmoid(result3)

    print ("--- Result: %f ..." %(result4))

    print("--- ");
    print("--- Part 6.3: Verify input [0,0,1,1] --> [0 or 1] ... ");
    print("--- ");

    single_point = np.array([0,0,1,1]) #1st step:

    result1 = np.dot(single_point, weight_hidden) 
    result2 = sigmoid(result1)
    result3 = np.dot(result2, weight_output)
    result4 = sigmoid(result3)

    print ("--- Result: %f ..." %(result4))

    print("--- ");
    print("--- Part 6.4: Verify input [0,0,0,1] --> 0 ... ");
    print("--- ");

    single_point = np.array([0,0,0,1]) #1st step:

    result1 = np.dot(single_point, weight_hidden) 
    result2 = sigmoid(result1)
    result3 = np.dot(result2, weight_output)
    result4 = sigmoid(result3)

    print ("--- Result: %f ..." %(result4))

    print("--- ============================================= ... ");
    print("--- Finished TestMultiLayer-Virus-Contraction.main() ... ");

# call the main method ...

main()
